// XML2IgorPro
// 2007.08.25
// J. J. Weimer
//
// Procedure files to parse an XML file and store it
//
// The XML file must have an XSL file containing definitions for
// XSL:TEMPLATE MATCH="//string": the name of each item for string wave storage
// VALUE-OF SELECT="string": the name of the value a given item
//
// The XSL style values are stored in root:Packages:XML2IgorPro:filenameXSL
// ename: string wave of SELECT values
// sname: string wave of STORE values
//
// Access to Package
//
// -> udFileLoaderXML(udFL)
//		STRUCTURE udFLStandardStructure &udFL
//
// see the structure definition and User's Manual

#pragma version=2.12
#pragma rtGlobals=1
#pragma IgorVersion=6.0

#include ":udStFiLrXML.lcl"

#include "udFLStandardStructure"
#include "udFLStandardFunctions"

// **********************************************
// Globals and Localizations

Static Structure xml2igorproglobals
	string version
	variable xslversionsupported
EndStructure

Static Function Initxml2igorproglobals(xip)
	STRUCT xml2igorproglobals &xip

	xip.version="2.12"
	xip.xslversionsupported=1.03
End

Static Function udFLXMLInititalize(udFL)
	STRUCT udFLStandardStructure &udFL
	
	udFLZeroStructure(udFL)
	udFLXMLQuery(udFL)
end

Static Function udFLXMLQuery(udFL)
	STRUCT udFLStandardStructure &udFL
	
	udFL.mimetype = "TEXT"
	udFL.extensions = ".xsl;.xml;"
	udFL.procModes = "Auto;Auto XML;Manual;"
	udFL.itemsDim[p] = p*0
	udFL.setNames=1
	
	switch(udFL.itemsType)
		case 2:
			udFL.itemsDim[0]=1		// keyword = tag
			break
		case 4:
			udFL.itemsDim[0] = 0		// text or numeric waves set by XSL?
			break
		case 8:
			udFL.itemsDim[0] = -1		// unscaled text
			udFL.itemsDim[1] = -1		// unscaled text
			udFL.itemsDim[2] = -1		// unscaled text
			break
		default:
			udFL.itemsType = 14		// mixture of returns
	endswitch
	
	return 0
	
end

// **********************************************
// File Loader

Function udStFiLrXML(udFL)
	STRUCT udFLStandardStructure &udFL

	variable refNum
	string pathStr, fileStr
	
	if (udFL.eventCode == -1)
		SetDataFolder root:
		KillDataFolder/Z root:Packages:XML2igorpro
		return 0
	endif
	
	switch(udFL.userCtrl)
		case -1:		// initialize		
			if (!DataFolderExists("root:Packages:xml2igorpro"))
				initParseXML()
			endif
			udFLXMLInititalize(udFL)		
			return 0
		case 0:		// query
			udFLXMLQuery(udFL)
			return 0
		case 1:
			if (strlen(udFL.fileList)==0)
				udFLSetError(udFL,alert0a,-1)
				return -1
			endif
			if (udFL.reportCtrl == 1)
				print "... confirming list of file names"
			endif
			confirmXMLFileList(udFL)
			if (udFL.reportCtrl == 1)
				print "... assuring file paths"
			endif
			if (udFLAssureFilePaths(udFL)<0)
				return -1
			endif
			break
		case 2:		// find XSL, use the name to find XML
			if (strlen(udFL.fileList) == 0)
				Open/D/R/M=xmldistr00/T=".xsl" refNum
				if (strlen(S_filename)!=0)
					udFL.filelist = S_filename + ";"
					udFL.fileList += ParseFilePath(1, S_filename, ":", 1, 0) + ParseFilePath(3, S_filename, ":", 0, 0) + ".xml;"
				else
					udFLSetError(udFL,alert06,-1)
					return -1
				endif
			endif
			if (udFL.reportCtrl == 1)
				print "... assuring file paths"
			endif
			if (udFLAssureFilePaths(udFL)<0)
				return -1
			endif
			if (udFL.reportCtrl == 1)
				print "... searching for XML"
			endif
			pathStr = udFL.pathStr
			NewPath/O/Q tmpXMLDir, pathStr
			fileStr = StringFromList(1,udFL.fileList)
			Open/Z=1/R/P=tmpXMLDir refNum as fileStr
			if (V_flag!=0)
				udFLSetError(udFL,alert05,-1)
				KillPath/Z tmpXMLDir
				return -1
			endif
			break
		case 3:		// find both XSL and XML
			Open/D/R/M=xmldistr00/T=".xsl" refNum
			if (strlen(S_filename)!=0)
				udFL.fileList = S_filename + ";"
			else
				udFLSetError(udFL,alert05,-1)
				return -1
			endif
			pathStr = ParseFilePath(1,StringFromList(0,udFL.fileList),":",1,0)
			NewPath/O/Q tmpXMLDir, pathStr
			Open/D/R/M=xmldistr01/T=".xml"/P=tmpXMLDir refNum
			if (strlen(S_filename)!=0)
				udFL.fileList += S_filename + ";"
				udFL.pathStr = ""
			else
				udFLSetError(udFL,alert05,-1)
				KillPath/Z tmpXMLDir
				return -1
			endif
			if (udFL.reportCtrl == 1)
				print "... assuring file paths"
			endif
			if (udFLAssureFilePaths(udFL)<0)
				return -1
			endif
			break
	endswitch
	if (udFL.reportCtrl == 1)
		print "... reading XSL"
	endif
	if (readXSLFile(udFL)<0)
		KillPath/Z tmpXMLDir
		return -1
	endif
	if (udFL.reportCtrl == 1)
		print "... parsing XSL"
	endif
	parseXSLFile(udFL)
	if (udFL.reportCtrl == 1)
		print "... reading XML"
	endif
	if (readXMLFile(udFL)<0)
		KillPath/Z tmpXMLDir
		return -1
	endif
	if (udFL.reportCtrl == 1)
		print "... parsing XML"
	endif
	if (parseXMLFile(udFL)<0)
		KillPath/Z tmpXMLDir
		return -1
	endif
	if (udFL.reportCtrl == 1)
		print "... final udFL", udFL
	endif
	if (udFL.reportCtrl >= 0)
		print "... loaded data ", udFL.itemsList
	endif
	KillPath/Z tmpXMLDir
	return 0
end

// **********************************************
// ENTRY POINTS TO FUNCTIONS ARE ABOVE THIS LINE
// Functions BELOW this point are SPECIFIC to this PACKAGE
// If you want to use them, change their names!!!!

// confirmXMLFileList(udFL)
// makes sure first file is .xsl and others are .xml during auto mode

Static Function confirmXMLFileList(udFL)
	STRUCT udFLStandardStructure &udFL

	string flist = udFL.fileList, tmpi, tmpo
	
	tmpi = ListMatch(flist,"*.xsl")
	tmpo = RemoveFromList(tmpi,flist)
	if (strlen(ListMatch(tmpo,"*.xsl"))!=0)
		udFLSetError(udFL,alert09,-1)
		return -1
	endif
	
	udFL.fileList = tmpi + tmpo
	
	return 0	
end

// initParseXML()
// create package folders
// return: 0-already existed; 1-newly created

Static Function initParseXML()

	string cdf=GetDataFolder(1)
	variable rc = 0

	NewDataFolder/O/S root:Packages
	if (!DataFolderExists("root:Packages:XML2IgorPro"))
		NewDataFolder/S XML2IgorPro
		string/G currXMLData, currXSLData
		rc = 1
	endif
	
	SetDataFolder $cdf
	
	return rc
end

// **********************************************
// readXSLFile(udFL)
// read the XML style sheet file as a notebook
// return: -1-unsuccessful; 0-successful

Static Function readXSLFile(udFL)
	STRUCT udFLStandardStructure &udFL
	
	variable refNum
	string fname, pathStr

	fname = StringFromList(0,udFL.fileList)
	if (strlen(udFL.pathStr)!=0)
		pathStr = udFL.pathStr	
		NewPath/O/Q tmpXMLDir, pathStr
		if (strlen(WinLIst("ParseXSLNB",";","WIN:16"))!=0)
			KillWindow ParseXSLNB
		endif
		OpenNotebook/Z/R/V=0/k=1/N=ParseXSLNB/P=tmpXMLDir fname
	else
		OpenNotebook/Z/R/V=0/k=1/N=ParseXSLNB fname
	endif
	if (V_flag!=0)
		udFLSetError(udFL,alert06b,-1)
		KillPath/Z tmpXMLDir
		return -1
	endif
	KillPath/Z tmpXMLDir
	udFL.itemsType = 16
	udFL.itemsCount = 1
	udFL.itemsList = "ParseXSLNB;"
	return 0
end

// **********************************************
// parseXSLFile(udFL)
// parse the XML style sheet file from notebook
// return: number of style elements found in file

Static Function parseXSLFile(udFL)
	STRUCT udFLStandardStructure &udFL

	SVAR currXSLData = root:Packages:XML2IgorPro:currXSLData

	STRUCT xml2igorproglobals xip
	Initxml2igorproglobals(xip)
	variable xslversionread
	
	string cdf=GetDataFolder(1)
	
	string xmlversion="igorpro version="
	string where
	
	where = ReplaceString(" ",ParseFilePath(3,StringFromList(0,udFL.fileList),":",0,0),"")
	
	string df = "root:Packages:XML2IgorPro:" + where +"XSL"

	NewDataFolder/O/S $df	
	currXSLData = df
	
	// ename holds element name as given in the database
	// sname holds symbolic representation of that name as the user desires it
	// scmd holds the command needed to generate the value
	// sunits holds the units for the value
	
	string/G eKey, firstKey
	make/o/n=1/T ename, sname, scmd, sunits
	
	variable spo, spo1, spf, spf1, ic, theValue, how
	
	string nb = "ParseXSLNB"
	string fstr, rstr, qstr = "\"", sstr = ";", estr = "/>"
	
	// search for XML version information
	
	Notebook $nb selection={startofFile,startofFile}, findText={xmlversion,0}
	if (!V_flag)
		DoAlert 0, alert00
		SetDataFolder $cdf	
		KillWindow $nb
		return 0
	endif
	Notebook $nb selection={startOfParagraph, endOfChars}
	GetSelection notebook, $nb, 2
	spo=strsearch(S_selection,xmlversion,0)
	spo=strsearch(S_selection,qstr,spo,2)+1
	spf=strsearch(S_selection,qstr,spo,2)-1
	xslversionread=str2num(S_selection[spo,spf])
	if (xslversionread>xip.xslversionsupported)
		DoAlert 0, alert01
		SetDataFolder $cdf	
		KillWindow $nb
		print alert01a, xip.xslversionsupported, xslversionread
		return 0
	endif
		
	// search for XML processing directive
	
	spo=strsearch(S_selection,"input-as=",spf,2)
	if (spo<0)
		udFLSetError(udFL,alert02,-1)
		SetDataFolder $cdf	
		KillWindow $nb
		return -1
	endif
	
	spo1 =strsearch(S_selection,qstr,spo,2)+1
	spf =strsearch(S_selection,qstr,spo1,2)-1
	fstr = S_selection[spo1,spf]
	strswitch(fstr)
		case "keyword-list":
			how=1
			udFL.itemsType = 2
			break
		case "waves":
			how=2
			udFL.itemsType = 4
			break
		case "row-waves":
			how=-2
			udFL.itemsType = -4
			break
		case "text-matrix":
			how=3
			udFL.itemsType = 8
			break
		default:
			DoAlert 1, alert03
			if (V_flag==1)
				how = 1
				udFL.itemsType = 2
				break
			else
				udFLSetError(udFL,alert03,-1)
				SetDataFolder $cdf
				KillWindow $nb
				return -1
			endif
	endswitch
	
	// search lines for MATCH code
	
	Notebook $nb selection={startofFile,startofFile}, findText={"match=",0}
	if (!V_flag)
		udFLSetError(udFL,alert04,-1)
		SetDataFolder $cdf	
		KillWindow $nb
		return -1
	endif
	
	Notebook $nb selection={startOfParagraph, endOfChars}
	GetSelection notebook, $nb, 2
	spo=strsearch(S_selection,"//",0)
	spf =strsearch(S_selection,">",spo,2)-2
	fstr = S_selection[spo+2,spf]
	
	// set the element KEY for MATCH=
	
	eKey = fstr
	
	//print fstr
	//return 0
	
	// search lines for value-of codes
	
	ic = 0
	
	do

		Notebook $nb selection={startofNextParagraph,startofNextParagraph}, findText={"value-of",0}
		if (!V_flag)
			break
		endif
		
		Redimension/N=(ic+1) ename, sname, sunits, scmd

		Notebook $nb selection={startOfParagraph, endOfChars}
		GetSelection notebook, $nb, 2
		
		//print ic, S_selection
	
		// select code
		
		spo1 = strsearch(S_selection,"select=",0,2)
		spf =strsearch(S_selection,qstr,spo1+8,2)-1
		fstr = S_selection[spo1+8,spf]
		
		ename[ic]=fstr
		
		if (ic == 0)
			firstKey = fstr
		endif
	
		// store code
		
		spo = strsearch(S_selection,"<xsl:value-of",0,2)
		spo1 = strsearch(S_selection,"=",spo,3)
		if (spo1>0)
			fstr = S_selection[0,spo1-1]
			spo1+=1
		else
			spo1 = 0
		endif
		
		sname[ic] = fstr
		
		// find everything between < and />
		
		spf = strsearch(S_selection,estr,0,2)+1
		rstr = S_selection[spo,spf]
		
		// find everything except the units
		
		spf1 = strlen(S_selection)
		if (spf1 > spf)
			spf = strsearch(S_selection,sstr,spf,2)-1
			if (spf < 0)
				spf = spf1
			endif
		endif
		fstr = S_selection[spo1,spf]
		
		scmd[ic] = ReplaceString(rstr,fstr,"theValue")
		
		// sunits
		
		spo1 = strsearch(S_selection,sstr,0,2)
		if (spo1>0)
			spo1 =strsearch(S_selection,qstr,spo1,2)
			if (spo1 > 0)
				spf =strsearch(S_selection,qstr,spo1+1,2)-1
				fstr = S_selection[spo1+1,spf]
				sunits[ic] = fstr
			endif
		endif
		
		//print "ename, sname, scmd, sunits: ", ename[ic], sname[ic], scmd[ic], sunits[ic]
		ic +=1
	while(1)
	
	SetDataFolder $cdf
	if (udFL.reportCtrl<1)
		KillWindow $nb
	endif
	
	udFL.itemsCount = 0
	return 0

end

// **********************************************
// readXMLFile(udFL)
// read the XML file as a notebook
// return: 0-unsuccessful; 1-successful

Function readXMLFile(udFL)
	STRUCT udFLStandardStructure &udFL
	
//	SVAR currXMLData = root:Packages:XML2IgorPro:currXMLData

	variable refNum
	string fname, pathStr
	
	fname = StringFromList(1,udFL.fileList)
	if (strlen(udFL.pathStr)!=0)
		pathStr = udFL.pathStr	
		NewPath/O/Q tmpXMLDir, pathStr
		if (strlen(WinLIst("ParseXMLNB",";","WIN:16"))!=0)
			KillWindow ParseXMLNB
		endif
		OpenNotebook/Z/R/V=0/k=1/N=ParseXMLNB/P=tmpXMLDir fname
	else
		OpenNotebook/Z/R/V=0/k=1/N=ParseXMLNB fname
	endif
	if (V_flag!=0)
		udFLSetError(udFL,alert06c,-1)
		return -1
	endif
	return 0

end

// **********************************************
// parseXMLFile(udFL)
//
// itemsType ... defines how
// 4: matrix
// 2: waves
// 1: keyword=value,units
// 0: test only
// toFolder ... defines xmlf where parsed data is stored
// fileList[0] ... defines folder string for xslf
// userData ... defines [oKey] as string
// 

Static Function parseXMLFile(udFL)
	STRUCT udFLStandardStructure &udFL
	
	SVAR currXSLData = root:Packages:XML2IgorPro:currXSLData
	SVAR currXMLData = root:Packages:XML2IgorPro:currXMLData
	NVAR nfmax=root:Packages:XML2IgorPro:nfieldsmax
	NVAR nemax=root:Packages:XML2IgorPro:nelementsmax
	
	nemax=0
	
	string where, xslf, oKey
	where = ReplaceString(" ",ParseFilePath(3,StringFromList(0,udFL.fileList),":",0,0),"")	
	xslf = "root:Packages:XML2IgorPro:" + where +"XSL"

	if (!DataFolderExists(xslf))
		DoAlert 0, alert06
		return 0
	endif
	
	// set data folder
	
	string cdf = GetDataFolder(1)
	
	currXSLData = "Packages:XML2IgorPro:" + where +"XSL"
	currXMLData = udFL.toFolder
	
	udFL.toFolder = currXSLData
	udFLSetDataFolder(udFL)
	
	wave/T ename = ename
	wave/T sname = sname
	wave/T scmd = scmd
	wave/T sunits = sunits
	SVAR eKey = eKey
	SVAR firstKey = firstKey
	
	if (strlen(udFL.userData)!=0)
		oKey = udFL.userData
	else
		oKEY=firstKey
	endif
	
	udFL.toFolder = currXMLData
	udFLSetDataFolder(udFL)
	
	string nb = "ParseXMLNB"	, theCmd, theSValue, theSUnits
	string eKstr, eKstrf, fstr, estr, tstr
	
	variable spo, spf, spo1, itempari, itemparf, nullValue=0
	variable/D dv
	variable/G theValue
	
	variable jc, ic, jmax, iemax, itmax, tc
	
	jmax = numpnts(ename)
		
	// search starting lines for first eKey code
	// (search for ELEMENT)

	eKstr = "<" + eKey +">"
	eKstrf = "</" + eKey + ">"
	//print eKstr
	Notebook $nb selection={startOfFile, startofFile}, findText={eKstr,0}
	if (!V_flag)
		DoAlert 0, alert07
		SetDataFolder $cdf	
		KillWindow $nb
		return 0
	endif
//	GetSelection notebook, $nb, 1
//	spo = V_startParagraph
	
	//return 0
	
	// count number of items
	
	Notebook $nb selection={startOfFile, endOfFile}
	GetSelection notebook, $nb, 1
	spf = V_endParagraph
	
	ic = 0
	jc = 0
	Notebook $nb selection={startOfFile, startofFile}
	do
		Notebook $nb selection={startOfNextParagraph, startofNextParagraph}, findText={eKstrf,0}
		if (V_flag)
			GetSelection notebook, $nb, 1
			ic += 1
		else
			Notebook $nb selection={startofParagraph,startofParagraph}, findText={eKstrf,0}
			GetSelection notebook, $nb, 1
		endif
	while(V_startParagraph<spf)
	
	// maximum number of items
	
	iemax = ic
	
	udFL.itemsList = ""
	udFL.itemsCount = 0
	
	//print iemax, spf	
	//return 0
	
	// loop over items starting at top
	// ic counts number of elements in XML
	// jc counts number of values per element (from XSL)
	
	ic = 0
	Notebook $nb selection={startOfFile, startofFile}
	
	do
		
		// search for eKey code for item ic
		// (search for ELEMENT again)
		
		eKstr = "<" + eKey
		Notebook $nb selection={startOfNextParagraph, startOfNextParagraph}, findText={eKstr,0}
		if (!V_flag)
			udFLSetError(udFL,alert07,-1)
			SetDataFolder $cdf	
			KillWindow $nb
			return -1
		endif
		
		GetSelection notebook, $nb, 3

		// store current location to return to for values search

		itempari = V_startParagraph
	
		// retrieve oKEY of item

		Notebook $nb selection={(itempari,0), (itempari,0)}, findText={"<"+oKEY,0}
		if (!V_flag)
			udFLSetError(udFL,alert07,-1)
			SetDataFolder $cdf
			KillWindow $nb
			return -1
		endif
		
		Notebook $nb selection={startOfParagraph, endOfChars}
		GetSelection notebook, $nb, 3
		spo1 = V_startParagraph
		spf =strsearch(S_selection,"</"+oKey+">",0,2)-1
		spo = strsearch(S_selection,">",spf,3)
		
		//print ic, iemax, S_selection, V_startParagraph
		//return 0
		
		estr = S_selection[spo+1,spf]

		//print spo, spf, estr
		//return 0
		
		// set the item value in the folder for keyword=list strings
		
		if (udFL.itemsType==1)
			string/G $estr = ""
			SVAR element = $estr
		endif
			
		// find the end limit of this search

		eKstrf = "</" + eKey + ">"
		Notebook $nb selection={(itempari,0),(itempari,0)}, findText={eKstrf,0}
		GetSelection notebook, $nb, 1
		itemparf = V_startParagraph
		
		// loop over VALUES for item (can be found in ANY order!)
			
		jc = 0
		do	
			
			// find the jc'th VALUE start code
			
			eKstr = "<" + ename[jc]
			eKstrf = "</" + ename[jc] + ">"
			
			//print "eKstrf: ", eKstrf
	
			Notebook $nb selection={(itempari,0),(itempari,0)}, findText={eKstr,0}
			GetSelection notebook, $nb, 1
			if (V_startParagraph>itemparf)
				nullValue = 1
				//print "no value: ", ic, fstr, eKstr
			else
				nullValue = 0
				Notebook $nb selection={startofParagraph,endofChars}
				GetSelection notebook, $nb, 3
				spo = V_startParagraph
			
				// find the jc'th VALUE end code
		
				Notebook $nb selection={startOfParagraph,startOfParagraph}, findText={eKstrf,0}
				if (!V_flag)
					DoAlert 0, alert08
					SetDataFolder $cdf	
					KillWindow $nb
					return 0
				endif
				Notebook $nb selection={StartofParagraph,endofChars}
				GetSelection notebook, $nb, 3
				spf = V_startParagraph + 1
				
				Notebook $nb selection={(spo,0),(spf,0)}
				GetSelection notebook, $nb, 2	
			endif
			
			switch(udFL.itemsType)
			case -4:   // user wants unscaled waves from row data
				if (nullValue==0)
					fstr = ParseXMLValue(S_selection,3)
					//print "the string for waves: ", fstr
					if (ic == 0)
						//print "the first string: ", fstr
						if (!GrepString(fstr,"(?i)[a-z]+"))
							// numeric value
							itmax = ItemsInList(fstr)
							make/o/n=(itmax) $sname[jc]=NaN
							if (strlen(sunits[jc])==0)
								fstr = ParseXMLValue(S_selection,0)
								if (cmpstr(fstr,"false")!=0)
									sunits[jc]=fstr
								endif
							endif
							if (strlen(sunits[jc])!=0)
								fstr = sname[jc]+"UNITS"
								string/G $fstr = sunits[jc]
							endif
						else
							// non-numeric value
							make/o/T/n=(itmax) $sname[jc]=""
						endif
					endif
					fstr = ParseXMLValue(S_selection,3)
					if (!GrepString(fstr,"(?i)[a-z]+"))
						// numeric value
						for (tc=0;tc<itmax;tc+=1)
							tstr = StringFromList(tc,fstr)
							theCmd = "theValue=" + ReplaceString("theValue",scmd[jc],tstr)
							//print "the command here is: ", theCmd
							Execute theCmd
							wave theWave = $sname[jc]
							theWave[tc] = theValue
						endfor
					else
						// non-numeric value
						for (tc=0;tc<itmax;tc+=1)
							wave/T theTWave = $sname[jc]
							fstr = ParseXMLValue(S_selection,2)
							theTWave[tc] = StringFromList(tc,fstr)
						endfor
					endif
					udFL.itemsList += sname[jc] + ";"
					udFL.itemsCount += 1
				endif
				break
			case 2:		// user wants keyword=value, units; list
				if (nullValue==0)
					if (ic == 0)
						if (strlen(sunits[jc])==0)
							fstr = ParseXMLValue(S_selection,0)
							if (cmpstr(fstr,"false")!=0)
								sunits[jc]=fstr
							endif
						endif
					endif
					fstr = ParseXMLValue(S_selection,1)
					if (!GrepString(fstr,"(?i)[a-z]+"))
						// numeric value
						theCmd = "theValue=" + ReplaceString("theValue",scmd[jc],fstr)
						Execute theCmd
						fstr = sname[jc] + "=" + num2str(theValue)
					else
						// non-numeric value
						fstr = sname[jc] + "=" + ParseXMLValue(S_selection,2)
					endif
				else
					fstr=sname[jc] + "="
				endif
				if (strlen(sunits[jc])>0)
					element += fstr + "," + sunits[jc] + ";"
				else
					element += fstr + ";"
				endif
				udFL.itemsList += estr + ";"
				udFL.itemsCount += 1
				break
			case 4:   // user wants unscaled waves
				if (nullValue==0)
					fstr = ParseXMLValue(S_selection,1)
					//print "the string for waves: ", fstr
					if (ic == 0)
						//print "the first string: ", fstr
						if (!GrepString(fstr,"(?i)[a-z]+"))
							// numeric value
							make/o/n=(iemax) $sname[jc]=NaN
							if (strlen(sunits[jc])==0)
								fstr = ParseXMLValue(S_selection,0)
								if (cmpstr(fstr,"false")!=0)
									sunits[jc]=fstr
								endif
							endif
							if (strlen(sunits[jc])!=0)
								fstr = sname[jc]+"UNITS"
								string/G $fstr = sunits[jc]
							endif
						else
							// non-numeric value
							make/o/T/n=(iemax) $sname[jc]=""
						endif
						fstr = ParseXMLValue(S_selection,1)
					endif
					if (!GrepString(fstr,"(?i)[a-z]+"))
						// numeric value
						theCmd = "theValue=" + ReplaceString("theValue",scmd[jc],fstr)
						//print "the command here is: ", theCmd
						Execute theCmd
						wave theWave = $sname[jc]
						theWave[ic] = theValue
					else
						// non-numeric value
						wave/T theTWave = $sname[jc]
						fstr = ParseXMLValue(S_selection,2)
						theTWave[ic] = fstr
					endif
				endif
				udFL.itemsList += sname[jc] + ";"
				udFL.itemsCount += 1
				break
			case 8:   // user wants unscaled 3-D text matrix
				if (nullValue==0)
					if ((ic == 0)&&(jc==0))
						make/T/o/n=(iemax,jmax,3) $sname[0]=""
						wave/T theTWave = $sname[0]
						theTWave[][0][1] = "value"
						theTWave[][0][2] = "units"
					endif
					fstr = ParseXMLValue(S_selection,1)
					//print "the string for matrix: ", fstr
					if (!GrepString(fstr,"(?i)[a-z]+"))
						// numeric value
						theCmd = "theValue=" + ReplaceString("theValue",scmd[jc],fstr)
						//print "the command here is: ", theCmd
						Execute theCmd
						theSValue = num2str(theValue)
						if (strlen(sunits[jc])==0)
							fstr = ParseXMLValue(S_selection,0)
							if (cmpstr(fstr,"false")!=0)
								sunits[jc]=fstr
							endif
						endif
						if (strlen(sunits[jc])!=0)
							theSUnits = sunits[jc]
						else
							theSUnits = ""
						endif
					else
						// non-numeric value
						fstr = ParseXMLValue(S_selection,2)
						theSValue = fstr
						theSUnits = ""
					endif
					wave/T theTWave = $sname[0]
					if (jc==0)
						theTWave[ic][0][0] = theSValue
					else
						theTWave[ic][jc][0] = sname[jc]
						theTWave[ic][jc][1] = theSValue
						theTWave[ic][jc][2] = theSUnits
					endif
					udFL.itemsList += sname[jc] + ";"
					udFL.itemsCount += 1
				endif
				//return 0
			endswitch			

//				print "the setup line: ", scmd[jc], sname[jc], fstr		
			
			//return 0
			
			jc += 1
			
		while(jc<jmax)	
		
		//return 0
		
		// move to next ELEMENT
		
		eKstrf = "</" + eKey + ">"
		Notebook $nb selection={(itempari,0),(itempari,0)}, findText={eKstrf,0}
		ic += 1
		
	while(ic<iemax)
	
	killvariables/Z theValue	
	SetDataFolder $cdf	
	KillWindow $nb
	
	return iemax

end

// **********************************************
// parseXMLValue(lstr,xhow)
// lstr: string (line) containing value to parse
// assumed to be in possible format ...
// <name UNITS=... bla bla bla><!--possible comments>value</name>
// including \r values ...
// returns: xhow=0 (units) xhow=1 (value with collapsed spaces)
//    xhow=2 (value with spaces) xhow=3 (value with spaces as ";") 

Static Function/S parseXMLValue(lstr,xhow)
	variable xhow
	string lstr
	
	string rstr
	variable so, sf
	lstr = ReplaceString("\r",lstr,"")
	
//	print "parsing string ", lstr
	
	switch(xhow)
		case 0:
			so = strsearch(lstr,"units=",0,2)
			if (so<0)
//				print "no units found"
				return "false"
			endif
			sf = strsearch(lstr,">",so+6)
			rstr = lstr[so+7,sf-2]
			break
		case 1:		
			sf = strsearch(lstr,"</",0)
			so = strsearch(lstr,">",sf,3)
			rstr = lstr[so+1,sf-1]
			rstr = ReplaceString(" ",rstr,"")
			break
		case 2:		
			sf = strsearch(lstr,"</",0)
			so = strsearch(lstr,">",sf,3)
			rstr = lstr[so+1,sf-1]
			break		
		case 3:		
			sf = strsearch(lstr,"</",0)
			so = strsearch(lstr,">",sf,3)
			rstr = lstr[so+1,sf-1]
			rstr = ReplaceString(" ",rstr,";")
			break
	endswitch
	
//	print "returning ", xhow, rstr
	return rstr
end